"""ext ctrl plns

Revision ID: 944a8fd5d5bc
Revises: 27724e407e2b
Create Date: 2025-09-10 16:18:19.742118

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '944a8fd5d5bc'
down_revision: Union[str, Sequence[str], None] = '27724e407e2b'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('k8s_ext_control_planes',
    sa.Column('stack_fqdn', sa.String(length=253), nullable=False),
    sa.Column('zone', sa.String(length=253), nullable=True),
    sa.Column('name', sa.String(length=253), nullable=True),
    sa.Column('proxy_stack_fqdn', sa.String(length=253), nullable=True),
    sa.PrimaryKeyConstraint('stack_fqdn')
    )
    op.drop_table('k8s_ext_control_plane')
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('k8s_ext_control_plane',
    sa.Column('stack_fqdn', sa.VARCHAR(length=253), autoincrement=False, nullable=False),
    sa.Column('zone', sa.VARCHAR(length=253), autoincrement=False, nullable=True),
    sa.Column('name', sa.VARCHAR(length=253), autoincrement=False, nullable=True),
    sa.Column('proxy_stack_fqdn', sa.VARCHAR(length=253), autoincrement=False, nullable=True),
    sa.PrimaryKeyConstraint('stack_fqdn', name=op.f('k8s_ext_control_plane_pkey'))
    )
    op.drop_table('k8s_ext_control_planes')
    # ### end Alembic commands ###



"""init

Revision ID: 7dea8c4ee39f
Revises: 
Create Date: 2025-09-09 14:30:52.222907

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '7dea8c4ee39f'
down_revision: Union[str, Sequence[str], None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('bind_domains',
    sa.Column('domain', sa.String(), nullable=False),
    sa.Column('stack_fqdn', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('domain', 'stack_fqdn')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('bind_domains')
    # ### end Alembic commands ###



"""machine type refactor

Revision ID: 0ad803c51325
Revises: fdcb5aa33b76
Create Date: 2025-12-22 13:46:51.857049

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = '0ad803c51325'
down_revision: Union[str, Sequence[str], None] = 'fdcb5aa33b76'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###

    op.add_column('k8s_tcp_proxies', sa.Column('is_k8s', sa.Boolean(), nullable=False, server_default='true'))
    op.add_column('kea_reservations', sa.Column('machine_type', sa.String(length=50), nullable=False, server_default='k8s_worker'))
    
    # remove defaults
    op.alter_column('k8s_tcp_proxies', 'is_k8s', server_default=None)
    op.alter_column('kea_reservations', 'machine_type', server_default=None)

    # set master type
    op.execute("""
        UPDATE kea_reservations kr
        SET machine_type = 'k8s_master'
        FROM k8s_masters km
        WHERE kr.ip = km.ip
    """)

    op.drop_table('k8s_masters')
    op.drop_table('k8s_workers')

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('kea_reservations', 'machine_type')
    op.drop_column('k8s_tcp_proxies', 'is_k8s')
    op.create_table('k8s_workers',
    sa.Column('ip', postgresql.INET(), autoincrement=False, nullable=False),
    sa.Column('hostname', sa.VARCHAR(length=253), autoincrement=False, nullable=False),
    sa.Column('stack_fqdn', sa.VARCHAR(length=253), autoincrement=False, nullable=False),
    sa.Column('proxy_stack_fqdn', sa.VARCHAR(length=253), autoincrement=False, nullable=False),
    sa.PrimaryKeyConstraint('ip', name=op.f('k8s_workers_pkey'))
    )
    op.create_table('k8s_masters',
    sa.Column('ip', postgresql.INET(), autoincrement=False, nullable=False),
    sa.Column('hostname', sa.VARCHAR(length=253), autoincrement=False, nullable=False),
    sa.Column('stack_fqdn', sa.VARCHAR(length=253), autoincrement=False, nullable=False),
    sa.Column('proxy_stack_fqdn', sa.VARCHAR(length=253), autoincrement=False, nullable=False),
    sa.PrimaryKeyConstraint('ip', name=op.f('k8s_masters_pkey'))
    )
    # ### end Alembic commands ###



"""ingress generic

Revision ID: e60b9cc63413
Revises: 0ad803c51325
Create Date: 2025-12-22 21:29:59.987939

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'e60b9cc63413'
down_revision: Union[str, Sequence[str], None] = '0ad803c51325'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('k8s_ingress_rules', sa.Column('is_k8s', sa.Boolean(), nullable=False, server_default='true'))
    op.alter_column('k8s_ingress_rules', 'is_k8s', server_default=None) # default create as true since to this point its only for k8s
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('k8s_ingress_rules', 'is_k8s')
    # ### end Alembic commands ###



import os

import paramiko
import pve_cloud._version
import yaml


def raise_on_py_cloud_missmatch(proxmox_host):
    # dont raise in tdd
    if os.getenv("PYTEST_CURRENT_TEST"):
        return

    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(proxmox_host, username="root")

    # since we need root we cant use sftp and root via ssh is disabled
    _, stdout, _ = ssh.exec_command("cat /etc/pve/cloud/cluster_vars.yaml")

    cluster_vars = yaml.safe_load(stdout.read().decode("utf-8"))

    if cluster_vars["py_pve_cloud_version"] != pve_cloud._version.__version__:
        raise RuntimeError(
            f"Version missmatch! py_pve_cloud_version for cluster is {cluster_vars['py_pve_cloud_version']}, while you are using {pve_cloud._version.__version__}"
        )



"""migrate old

Revision ID: 7868bcd05006
Revises: 7dea8c4ee39f
Create Date: 2025-09-09 18:09:39.760976

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision: str = '7868bcd05006'
down_revision: Union[str, Sequence[str], None] = '7dea8c4ee39f'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('acme_x509',
    sa.Column('stack_fqdn', sa.String(length=253), nullable=False),
    sa.Column('config', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    sa.Column('ec_csr', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    sa.Column('ec_crt', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    sa.Column('k8s', postgresql.JSONB(astext_type=sa.Text()), nullable=True),
    sa.PrimaryKeyConstraint('stack_fqdn')
    )
    op.create_table('k8s_ingress_rules',
    sa.Column('zone', sa.String(length=253), nullable=False),
    sa.Column('name', sa.String(length=253), nullable=False),
    sa.Column('stack_fqdn', sa.String(length=253), nullable=False),
    sa.Column('proxy_stack_fqdn', sa.String(length=253), nullable=True),
    sa.Column('external', sa.Boolean(), nullable=True),
    sa.PrimaryKeyConstraint('zone', 'name', 'stack_fqdn')
    )
    op.create_table('k8s_masters',
    sa.Column('ip', postgresql.INET(), nullable=False),
    sa.Column('hostname', sa.String(length=253), nullable=True),
    sa.Column('stack_fqdn', sa.String(length=253), nullable=True),
    sa.Column('proxy_stack_fqdn', sa.String(length=253), nullable=True),
    sa.PrimaryKeyConstraint('ip')
    )
    op.create_table('k8s_tcp_proxies',
    sa.Column('proxy_name', sa.String(length=253), nullable=True),
    sa.Column('haproxy_port', sa.SmallInteger(), nullable=False),
    sa.Column('node_port', sa.SmallInteger(), nullable=True),
    sa.Column('stack_fqdn', sa.String(length=253), nullable=True),
    sa.Column('proxy_snippet', sa.Text(), nullable=True),
    sa.Column('proxy_stack_fqdn', sa.String(length=253), nullable=False),
    sa.Column('external', sa.Boolean(), server_default='false', nullable=True),
    sa.PrimaryKeyConstraint('haproxy_port', 'proxy_stack_fqdn')
    )
    op.create_table('k8s_workers',
    sa.Column('ip', postgresql.INET(), nullable=False),
    sa.Column('hostname', sa.String(length=253), nullable=True),
    sa.Column('stack_fqdn', sa.String(length=253), nullable=True),
    sa.Column('proxy_stack_fqdn', sa.String(length=253), nullable=True),
    sa.PrimaryKeyConstraint('ip')
    )
    op.create_table('kea_reservations',
    sa.Column('mac', postgresql.MACADDR(), nullable=False),
    sa.Column('ip', postgresql.INET(), nullable=False),
    sa.Column('hostname', sa.String(length=253), nullable=True),
    sa.Column('stack_fqdn', sa.String(length=253), nullable=True),
    sa.PrimaryKeyConstraint('ip')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('kea_reservations')
    op.drop_table('k8s_workers')
    op.drop_table('k8s_tcp_proxies')
    op.drop_table('k8s_masters')
    op.drop_table('k8s_ingress_rules')
    op.drop_table('acme_x509')
    # ### end Alembic commands ###



"""external cp extra sans

Revision ID: 04398db10434
Revises: 24a548bfce3e
Create Date: 2025-11-20 09:45:26.563717

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '04398db10434'
down_revision: Union[str, Sequence[str], None] = '24a548bfce3e'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('k8s_ext_control_planes', sa.Column('extra_sans', sa.String(length=2530), nullable=True))

    op.execute("""
        UPDATE k8s_ext_control_planes
        SET extra_sans = name || '.' || zone
    """)

    op.alter_column('k8s_ext_control_planes', 'extra_sans', nullable=False)

    op.drop_column('k8s_ext_control_planes', 'name')
    op.drop_column('k8s_ext_control_planes', 'zone')
    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade schema."""
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('k8s_ext_control_planes', sa.Column('zone', sa.VARCHAR(length=253), autoincrement=False, nullable=False))
    op.add_column('k8s_ext_control_planes', sa.Column('name', sa.VARCHAR(length=253), autoincrement=False, nullable=False))
    op.drop_column('k8s_ext_control_planes', 'extra_sans')
    # ### end Alembic commands ###



import os
import shutil
import socket
import subprocess

import yaml
from proxmoxer import ProxmoxAPI

from pve_cloud.lib.validate import raise_on_py_cloud_missmatch


def get_cloud_domain(target_pve, suppress_warnings=False):
    if shutil.which("avahi-browse"):
        avahi_disc = subprocess.run(
            ["avahi-browse", "-rpt", "_pxc._tcp"],
            stdout=subprocess.PIPE,
            text=True,
            check=True,
        )
        services = avahi_disc.stdout.splitlines()

        # find cloud domain hosts and get first online per proxmox cluster
        for service in services:
            if service.startswith("="):
                # avahi service def
                svc_args = service.split(";")

                cloud_domain = None
                cluster_name = None

                for txt_arg in svc_args[9].split():
                    txt_arg = txt_arg.replace('"', "")
                    if txt_arg.startswith("cloud_domain"):
                        cloud_domain = txt_arg.split("=")[1]

                    if txt_arg.startswith("cluster_name"):
                        cluster_name = txt_arg.split("=")[1]

                if not cloud_domain or not cluster_name:
                    raise ValueError(
                        f"Missconfigured proxmox cloud avahi service: {service}"
                    )

                if target_pve.endswith(cloud_domain):
                    return cloud_domain

        raise RuntimeError("Could not get cloud domain via avahi mdns!")
    else:
        if not suppress_warnings:
            print(
                "avahi-browse not available, falling back to local inventory file from pvcli connect-cluster!"
            )

        with open(os.path.expanduser("~/.pve-cloud-dyn-inv.yaml"), "r") as f:
            pve_inventory = yaml.safe_load(f)

        for pve_cloud in pve_inventory:
            for pve_cluster in pve_inventory[pve_cloud]:
                if pve_cluster + "." + pve_cloud == target_pve:
                    return pve_cloud

        raise Exception(f"Could not identify cloud domain for {target_pve}")


def get_online_pve_host(target_pve, suppress_warnings=False, skip_py_cloud_check=False):
    if shutil.which("avahi-browse"):
        avahi_disc = subprocess.run(
            ["avahi-browse", "-rpt", "_pxc._tcp"],
            stdout=subprocess.PIPE,
            text=True,
            check=True,
        )
        services = avahi_disc.stdout.splitlines()

        for service in services:
            if service.startswith("="):
                # avahi service def
                svc_args = service.split(";")
                host_ip = svc_args[7]

                cloud_domain = None
                cluster_name = None

                for txt_arg in svc_args[9].split():
                    txt_arg = txt_arg.replace('"', "")
                    if txt_arg.startswith("cloud_domain"):
                        cloud_domain = txt_arg.split("=")[1]

                    if txt_arg.startswith("cluster_name"):
                        cluster_name = txt_arg.split("=")[1]

                if not cloud_domain or not cluster_name:
                    raise ValueError(
                        f"Missconfigured proxmox cloud avahi service: {service}"
                    )

                # main pve cloud inventory
                if f"{cluster_name}.{cloud_domain}" == target_pve:
                    if not skip_py_cloud_check:
                        raise_on_py_cloud_missmatch(
                            host_ip
                        )  # validate that versions of dev machine and running on cluster match

                    return host_ip

        raise RuntimeError(f"No online host found for {target_pve}!")
    else:
        if not suppress_warnings:
            print(
                "avahi-browse not available, falling back to local inventory file from pvcli connect-cluster!"
            )

        with open(os.path.expanduser("~/.pve-cloud-dyn-inv.yaml"), "r") as f:
            pve_inventory = yaml.safe_load(f)

        for pve_cloud in pve_inventory:
            for pve_cluster in pve_inventory[pve_cloud]:
                if pve_cluster + "." + pve_cloud == target_pve:
                    for pve_host in pve_inventory[pve_cloud][pve_cluster]:
                        # check if host is available
                        pve_host_ip = pve_inventory[pve_cloud][pve_cluster][pve_host][
                            "ansible_host"
                        ]
                        try:
                            with socket.create_connection((pve_host_ip, 22), timeout=3):

                                if not skip_py_cloud_check:
                                    raise_on_py_cloud_missmatch(
                                        pve_host_ip
                                    )  # validate that versions of dev machine and running on cluster match

                                return pve_host_ip
                        except Exception as e:
                            # debug
                            print(e, type(e))

        raise RuntimeError(f"Could not find online pve host for {target_pve}")


def get_pve_inventory(
    pve_cloud_domain,
    suppress_warnings=False,
    skip_py_cloud_check=False,
    fetch_other_pve_hosts=False,
):
    if shutil.which("avahi-browse"):
        # avahi is available

        # call avahi-browse -rpt _pxc._tcp and find online host matching pve cloud domain
        # connect via ssh and fetch all other hosts via proxmox api => build inventory
        avahi_disc = subprocess.run(
            ["avahi-browse", "-rpt", "_pxc._tcp"],
            stdout=subprocess.PIPE,
            text=True,
            check=True,
        )
        services = avahi_disc.stdout.splitlines()

        pve_inventory = {}

        py_pve_cloud_performed_version_checks = set()

        # find cloud domain hosts and get first online per proxmox cluster
        cloud_domain_first_hosts = {}
        for service in services:
            if service.startswith("="):
                # avahi service def
                svc_args = service.split(";")
                host_name = svc_args[6].removesuffix(".local")
                host_ip = svc_args[7]

                cloud_domain = None
                cluster_name = None

                for txt_arg in svc_args[9].split():
                    txt_arg = txt_arg.replace('"', "")
                    if txt_arg.startswith("cloud_domain"):
                        cloud_domain = txt_arg.split("=")[1]

                    if txt_arg.startswith("cluster_name"):
                        cluster_name = txt_arg.split("=")[1]

                if not cloud_domain or not cluster_name:
                    raise ValueError(
                        f"Missconfigured proxmox cloud avahi service: {service}"
                    )

                # build inventory only for the current domain
                if cloud_domain == pve_cloud_domain:
                    if cluster_name not in pve_inventory:
                        pve_inventory[cluster_name] = {}

                    pve_inventory[cluster_name][host_name] = {
                        "ansible_user": "root",
                        "ansible_host": host_ip,
                    }

                # main pve cloud inventory
                if (
                    cloud_domain == pve_cloud_domain
                    and cluster_name not in cloud_domain_first_hosts
                ):
                    if (
                        not skip_py_cloud_check
                        and f"{cluster_name}.{cloud_domain}"
                        not in py_pve_cloud_performed_version_checks
                    ):
                        raise_on_py_cloud_missmatch(
                            host_ip
                        )  # validate that versions of dev machine and running on cluster match
                        py_pve_cloud_performed_version_checks.add(
                            f"{cluster_name}.{cloud_domain}"
                        )  # perform version check only once per cluster

                    cloud_domain_first_hosts[cluster_name] = host_ip

        if not fetch_other_pve_hosts:
            return pve_inventory  # return without doing inter api call resolution

        # iterate over hosts and build pve inv via proxmox api
        # todo: this needs to be hugely optimized it blocks the grpc server
        for cluster_first, first_host in cloud_domain_first_hosts.items():
            proxmox = ProxmoxAPI(first_host, user="root", backend="ssh_paramiko")

            cluster_name = None
            status_resp = proxmox.cluster.status.get()
            for entry in status_resp:
                if entry["id"] == "cluster":
                    cluster_name = entry["name"]
                break

            if cluster_name is None:
                raise RuntimeError("Could not get cluster name")

            if cluster_name != cluster_first:
                raise ValueError(
                    f"Proxmox cluster name missconfigured in avahi service {cluster_name}/{cluster_first}"
                )

            # fetch other hosts via api
            cluster_hosts = proxmox.nodes.get()

            for node in cluster_hosts:
                node_name = node["node"]

                if node["status"] == "offline":
                    print(f"skipping offline node {node_name}")
                    continue

                # get the main ip
                ifaces = proxmox.nodes(node_name).network.get()
                node_ip_address = None
                for iface in ifaces:
                    if "gateway" in iface:
                        if node_ip_address is not None:
                            raise RuntimeError(
                                f"found multiple ifaces with gateways for node {node_name}"
                            )
                        node_ip_address = iface.get("address")

                if node_ip_address is None:
                    raise RuntimeError(f"Could not find ip for node {node_name}")

                pve_inventory[cluster_name][node_name] = {
                    "ansible_user": "root",
                    "ansible_host": node_ip_address,
                }

        return pve_inventory

    else:
        if not suppress_warnings:
            print(
                "avahi-browse not available, falling back to local inventory file from pvcli connect-cluster!"
            )
        # try load fallback manual inventory from disk
        inv_path = os.path.expanduser("~/.pve-cloud-dyn-inv.yaml")
        if not os.path.exists(inv_path):
            raise RuntimeError(
                "Local pve inventory file missing (~/.pve-cloud-dyn-inv.yaml), execute `pvcli connect-cluster` or setup avahi mdns discovery!"
            )

        with open(inv_path, "r") as file:
            dynamic_inventory = yaml.safe_load(file)

        if pve_cloud_domain not in dynamic_inventory:
            raise RuntimeError(
                f"{pve_cloud_domain} not in local dynamic inventory (~/.pve-cloud-dyn-inv.yaml created by `pvcli connect-cluster`)!"
            )

        return dynamic_inventory[pve_cloud_domain]



import argparse
import os

import paramiko
import yaml
from proxmoxer import ProxmoxAPI

from pve_cloud.cli.pvclu import get_ssh_master_kubeconfig
from pve_cloud.lib.inventory import *


def connect_cluster(args):
    # try load current dynamic inventory
    inv_path = os.path.expanduser("~/.pve-cloud-dyn-inv.yaml")
    if os.path.exists(inv_path):
        with open(inv_path, "r") as file:
            dynamic_inventory = yaml.safe_load(file)
    else:
        # initialize empty
        dynamic_inventory = {}

    # connect to the cluster via paramiko and check if cloud files are already there
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    ssh.connect(args.pve_host, username="root")

    # since we need root we cant use sftp and root via ssh is disabled
    _, stdout, _ = ssh.exec_command("cat /etc/pve/cloud/cluster_vars.yaml")
    cluster_vars = yaml.safe_load(stdout.read().decode("utf-8"))

    if not cluster_vars:
        # cluster has not been yet initialized
        pve_cloud_domain = input(
            "Cluster has not yet been fully initialized, assign the cluster a cloud domain and press ENTER:"
        )
    else:
        pve_cloud_domain = cluster_vars["pve_cloud_domain"]

    # init cloud domain if not there
    if pve_cloud_domain not in dynamic_inventory:
        dynamic_inventory[pve_cloud_domain] = {}

    # connect to the passed host
    proxmox = ProxmoxAPI(args.pve_host, user="root", backend="ssh_paramiko")

    # try get the cluster name
    cluster_name = None
    status_resp = proxmox.cluster.status.get()
    for entry in status_resp:
        if entry["id"] == "cluster":
            cluster_name = entry["name"]
            break

    if cluster_name is None:
        raise Exception("Could not get cluster name")

    if cluster_name in dynamic_inventory[pve_cloud_domain] and not args.force:
        print(
            f"cluster {cluster_name} already in dynamic inventory, add --force to overwrite current local inv."
        )
        return

    # overwrite on force / create fresh
    dynamic_inventory[pve_cloud_domain][cluster_name] = {}

    # not present => add and safe the dynamic inventory
    cluster_hosts = proxmox.nodes.get()

    for node in cluster_hosts:
        node_name = node["node"]

        if node["status"] == "offline":
            print(f"skipping offline node {node_name}")
            continue

        # get the main ip
        ifaces = proxmox.nodes(node_name).network.get()
        node_ip_address = None
        for iface in ifaces:
            if "gateway" in iface:
                if node_ip_address is not None:
                    raise Exception(
                        f"found multiple ifaces with gateways for node {node_name}"
                    )
                node_ip_address = iface.get("address")

        if node_ip_address is None:
            raise Exception(f"Could not find ip for node {node_name}")

        print(f"adding {node_name}")
        dynamic_inventory[pve_cloud_domain][cluster_name][node_name] = {
            "ansible_user": "root",
            "ansible_host": node_ip_address,
        }

    print(f"writing dyn inv to {inv_path}")
    with open(inv_path, "w") as file:
        yaml.dump(dynamic_inventory, file)


def print_kubeconfig(args):
    if not os.path.exists(args.inventory):
        print("The specified inventory file does not exist!")
        return

    with open(args.inventory, "r") as f:
        inventory = yaml.safe_load(f)

    target_pve = inventory["target_pve"]

    target_cloud_domain = get_cloud_domain(target_pve)
    pve_inventory = get_pve_inventory(target_cloud_domain)

    # find target cluster in loaded inventory
    target_cluster = None

    for cluster in pve_inventory:
        if target_pve.endswith((cluster + "." + target_cloud_domain)):
            target_cluster = cluster
            break

    if not target_cluster:
        print("could not find target cluster in pve inventory!")
        return

    first_host = list(pve_inventory[target_cluster].keys())[0]

    # connect to the first pve host in the dyn inv, assumes they are all online
    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh.connect(
        pve_inventory[target_cluster][first_host]["ansible_host"], username="root"
    )

    # since we need root we cant use sftp and root via ssh is disabled
    _, stdout, _ = ssh.exec_command("cat /etc/pve/cloud/cluster_vars.yaml")

    cluster_vars = yaml.safe_load(stdout.read().decode("utf-8"))

    print(get_ssh_master_kubeconfig(cluster_vars, inventory["stack_name"]))


def main():
    parser = argparse.ArgumentParser(
        description="PVE general purpose cli for setting up."
    )

    base_parser = argparse.ArgumentParser(add_help=False)

    subparsers = parser.add_subparsers(dest="command", required=True)

    connect_cluster_parser = subparsers.add_parser(
        "connect-cluster",
        help="Add an entire pve cluster to this machine for use.",
        parents=[base_parser],
    )
    connect_cluster_parser.add_argument(
        "--pve-host",
        type=str,
        help="PVE Host to connect to and add the entire cluster for the local machine.",
        required=True,
    )
    connect_cluster_parser.add_argument(
        "--force", action="store_true", help="Will read the cluster if set."
    )
    connect_cluster_parser.set_defaults(func=connect_cluster)

    print_kconf_parser = subparsers.add_parser(
        "print-kubeconfig",
        help="Print the kubeconfig from a k8s cluster deployed with pve cloud.",
        parents=[base_parser],
    )
    print_kconf_parser.add_argument(
        "--inventory",
        type=str,
        help="PVE cloud kubespray inventory yaml file.",
        required=True,
    )
    print_kconf_parser.set_defaults(func=print_kubeconfig)

    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()
